from uuid import UUID
from datetime import datetime
import sys

# https://uuid6.github.io/uuid6-ietf-draft/
GREGORIAN_UNIX_OFFSET = 12219292800000000000

# https://uuid.ramsey.dev/en/stable/rfc4122/version2.html#lossy-timestamps
V2_CLOCK_TICK = 429.4967295 


def error(*args, **kwargs):
    """Print an error message and exit"""
    print(*args, **kwargs, file=sys.stderr)
    exit(1)

def is_uuid(uuid_str: str) -> bool:
    """Check if a string is a valid UUID

    Args:
        uuid_str (str): The string to check

    Returns:
        bool: True if the string is a valid UUID, False otherwise
    """
    uuid_str = uuid_str.replace("-", "")
    return len(uuid_str) == 32 and all(c in "0123456789abcdef" for c in uuid_str.lower())

def get_uuid(uuid_str: str) -> UUID:
    """Get a UUID from a string or exit if the string is not a valid UUID

    Args:
        uuid_str (str): The string to convert to a UUID

    Returns:
        UUID: The UUID
    """
    if not is_uuid(uuid_str):
        error(f"{uuid_str} is not a valid UUID")
    return UUID(uuid_str)

def get_version(uuid: UUID) -> int:
    """Get the version of a UUID

    Args:
        uuid (UUID): The UUID to get the version from

    Returns:
        int: The version of the UUID
    """
    return (uuid.int >> 76) & 0xf

def get_variant(uuid: UUID) -> int:
    """Get the variant of a UUID

    Args:
        uuid (UUID): The UUID to get the variant from

    Returns:
        int: The variant of the UUID
    """
    return (uuid.int  >> 60) & 0xf

def get_timestamp(uuid: UUID) -> int:
    """Get the timestamp from a UUID

    Args:
        uuid (UUID): The UUID to get the timestamp from

    Returns:
        int: The timestamp in nanoseconds
    """
    version = get_version(uuid)
    
    if version == 1:
        return (uuid.time * 100) - GREGORIAN_UNIX_OFFSET
    elif version == 2:
        timestamp_low = int.from_bytes(uuid.bytes[4:6])
        timestamp_high = int.from_bytes(uuid.bytes[6:8]) & 0x0fff
        return int(((timestamp_high << 16) | timestamp_low) * V2_CLOCK_TICK * 1e9) - GREGORIAN_UNIX_OFFSET
    elif version == 6:
        return ((uuid.int >> 80) << 12) + ((uuid.int >> 64) & 4095) * 100 - GREGORIAN_UNIX_OFFSET
    elif version == 7:
        return (uuid.int >> 80) * 1_000_000

    
def alt_sort(timestamps: list[int]) -> list[int]:
    """Sort a list of timestamps in an alternating pattern

    Args:
        timestamps (list[int]): The timestamps to sort

    Returns:
        list[int]: The sorted timestamps
    """
    out = []
    size = len(timestamps)
    if len(timestamps) // 2 != 0:
        idx = size // 2
        out.append(timestamps[idx])
        i1, i2 = idx - 1, idx + 1
    else:
        idx = size // 2
        i1, i2 = idx - 1, idx
        
    while i1 >= 0:
        out.append(timestamps[i1])
        if i2 < size:
            out.append(timestamps[i2])
        i1 -= 1
        i2 += 1
        
    return out

def parse_time(time_str: str) -> int:
    """Parse a string representing a time into an integer

    Args:
        time_str (str): The time string to parse

    Returns:
        int: The time in nanoseconds
    """
    
    if time_str is None:
        return None
    
    try:
        return int(datetime.fromisoformat(time_str).timestamp() * 1e9)
    except ValueError:
        pass
    
    try:
        time = float(time_str)
        if time > 0xffffffff:
            return int(time) 
        else:
            return int(time * 1e9)
    except ValueError:
        error("Time must be an integer or an ISO 8601 formatted string")
        
def get_int(arg: str, error_message: str, base: int = 10) -> int:
    """Util function tu get an integer from a string

    Args:
        arg (str): The string to convert to an integer
        error_message (str): The error message to display if the conversion fails
        base (int, optional): The base to use for the conversion. Defaults to 10.

    Returns:
        int: The integer
    """
    
    if arg is None:
        return None
    
    if base == 16:
        arg = arg.replace(":","")
    
    try:
        return int(arg, base)
    except ValueError:
        error(error_message)
